/*********************************************************************************************************************************
 * Copyright (c) Erik Kunze 1995 - 2003
 *
 * This file is a part of the Spectrum emulator package 'XZX-Pro', and may not be distributed.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Erik Kunze
 ********************************************************************************************************************************/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#ifndef lint
static char rcsid[]="$Id: joystick.c,v 4.25 2003/03/17 07:33:50 erik Rel $";
#endif
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#ifdef JOY
#include JOYSTICK_H_PATH
#endif
#include <assert.h>
#include "monitor/debug.h"
#include "ui/resource.h"
#include "machine/machine.h"
#include "util.h"
#include "emul.h"
#include "event.h"
#include "joystick.h"
#if defined(JS_VERSION) && JS_VERSION > 0x010000
#define NEW_INTERFACE
#endif
#if defined(__FreeBSD__) || defined(__NetBSD__)
#define JS_DATA_TYPE	joystick
#define JS_RETURN	sizeof(struct JS_DATA_TYPE)
#endif
#ifdef DEBUG
#define DEB(x)	{ if ((CFGGET(debug) & D_JOY) != 0) { x } }
#else
#define DEB(x)
#endif
#ifdef JOY
static void joyAutofireHandler(void);static void joyJuggleHandler(void);
#endif
JoyData*JoyDataP;void JoyInit(void){JoyDataP=Malloc(sizeof(JoyData),"JoyInit");
#ifdef JOY
if((JoyDataP->fd=open(CFGGET(joyDev),O_RDONLY|O_NONBLOCK))<1){Msg(M_PERR,"couldn't open joystick device for reading");CFGSET(joy,-1);return;}{
#ifdef NEW_INTERFACE
unsigned int version;char name[256];char axes=2,buttons=2;(void)ioctl(JoyDataP->fd,JSIOCGVERSION,&version);(void)ioctl(JoyDataP->fd,JSIOCGAXES,&axes);(void)ioctl(JoyDataP->fd,JSIOCGBUTTONS,&buttons);(void)ioctl(JoyDataP->fd,JSIOCGNAME(sizeof(name)),name);if(version>0x010000){JoyDataP->new=1;DEB(Msg(M_DEBUG,"JOY: driver version %d.%d.%d",version>>16,(version>>8)&0xff,version&0xff);Msg(M_DEBUG,"JOY: %s, %d axes, %d buttons",name,(int)axes,(int)buttons););if(buttons<3){CFGSET(joyJuggle,-CFGGET(joyJuggle));}if(buttons<2){CFGSET(joyAutofire,-CFGGET(joyAutofire));}}else
#endif
{struct JS_DATA_TYPE jdata;JoyDataP->new=0;DEB(Msg(M_DEBUG,"JOY: driver version 0.8.X");Msg(M_DEBUG,"JOY: unknown joystick type"););if(read(JoyDataP->fd,&jdata,JS_RETURN)!=JS_RETURN){Msg(M_PERR,"couldn't read current joystick settings");CFGSET(joy,-1);}else{JoyDataP->centreX=jdata.x;JoyDataP->centreY=jdata.y;}CFGSET(joyJuggle,-CFGGET(joyJuggle));}}
#endif
}
#ifdef JOY
void JoyRead(void){
#ifdef NEW_INTERFACE
if(JoyDataP->new!=0){struct js_event jev;if(read(JoyDataP->fd,&jev,sizeof(jev))==sizeof(jev)){if((jev.type&~JS_EVENT_INIT)==JS_EVENT_BUTTON){switch(jev.number){case 0:if(jev.value!=0){JOY_PRESS(fire);}else{JOY_RELEASE(fire);}break;case 1:if(jev.value!=0){if(JOY_PRESSED(fire)){JOY_RELEASE(autofire);}else{JOY_PRESS(autofire);joyAutofireHandler();}}else{JOY_RELEASE(autofire);JOY_RELEASE(fire);}break;case 2:if(jev.value!=0){if(JOY_PRESSED(left)||JOY_PRESSED(right)){JOY_RELEASE(juggle);}else{JOY_PRESS(juggle);JOY_PRESS(left);joyJuggleHandler();}}else{JOY_RELEASE(juggle);JOY_RELEASE(left);JOY_RELEASE(right);}break;case 4:if(jev.value!=0){if(JOY_PRESSED(autofire)){if(CFGGET(joyAutofire)>1){CFGSET(joyAutofire,CFGGET(joyAutofire)-1);}}else if(JOY_PRESSED(juggle)){if(CFGGET(joyJuggle)>1){CFGSET(joyJuggle,CFGGET(joyJuggle)-1);}}else if(CFGGET(fastmode)==0){unsigned int speed=CFGGET(speed);if(speed>10){speed-=10;CFGSET(speed,speed);EmulSetSpeed(speed);}}}break;case 5:if(jev.value!=0){if(JOY_PRESSED(autofire)){if(CFGGET(joyAutofire)<50){CFGSET(joyAutofire,CFGGET(joyAutofire)+1);}}else if(JOY_PRESSED(juggle)){if(CFGGET(joyJuggle)<50){CFGSET(joyJuggle,CFGGET(joyJuggle)+1);}}else if(CFGGET(fastmode)==0){unsigned int speed=CFGGET(speed);if(speed<=190){speed+=10;CFGSET(speed,speed);EmulSetSpeed(speed);}}}break;}}else{if((int)jev.number==CFGGET(joyAxisX)){if(jev.value<=-CFGGET(joyTolerance)){JOY_PRESS(left);}else if(jev.value>=CFGGET(joyTolerance)){JOY_PRESS(right);}else{JOY_RELEASE(left);JOY_RELEASE(right);}}else if((int)jev.number==CFGGET(joyAxisY)){if(jev.value<=-CFGGET(joyTolerance)){JOY_PRESS(up);}else if(jev.value>=CFGGET(joyTolerance)){JOY_PRESS(down);}else{JOY_RELEASE(up);JOY_RELEASE(down);}}}}}else
#endif
{struct JS_DATA_TYPE jdata;if(read(JoyDataP->fd,&jdata,JS_RETURN)!=JS_RETURN){Msg(M_PWARN,"lost joystick connection");}else{JOY_RELEASE(left);JOY_RELEASE(right);JOY_RELEASE(up);JOY_RELEASE(down);if(!JOY_PRESSED(autofire)){JOY_RELEASE(fire);}if(jdata.x<=JoyDataP->centreX-CFGGET(joyTolerance)){JOY_PRESS(left);}else if(jdata.x>=JoyDataP->centreX+CFGGET(joyTolerance)){JOY_PRESS(right);}if(jdata.y<=JoyDataP->centreY-CFGGET(joyTolerance)){JOY_PRESS(up);}else if(jdata.y>=JoyDataP->centreY+CFGGET(joyTolerance)){JOY_PRESS(down);}
#if defined(__FreeBSD__) || defined(__NetBSD__)
if(jdata.b1)
#else
if(jdata.buttons&0x01)
#endif
{JOY_PRESS(fire);}
#if defined(__FreeBSD__) || defined(__NetBSD__)
else if(jdata.b2)
#else
else if(jdata.buttons&0x02)
#endif
{if(!JOY_PRESSED(autofire)){JOY_PRESS(autofire);joyAutofireHandler();}}else{JOY_RELEASE(autofire);}}}}
#endif
#ifdef JOY
static void joyAutofireHandler(void){if(JOY_PRESSED(autofire)){JOY_TOGGLE(fire);EventAdd((EventFunc)joyAutofireHandler,(void*)0,Machine->tstatesPerFrame*50/CFGGET(joyAutofire));}}
#endif
#ifdef JOY
static void joyJuggleHandler(void){if(JOY_PRESSED(juggle)){JOY_TOGGLE(left);JOY_TOGGLE(right);EventAdd((EventFunc)joyJuggleHandler,(void*)0,Machine->tstatesPerFrame*50/CFGGET(joyJuggle));}}
#endif

