/*********************************************************************************************************************************
 * Copyright (c) Erik Kunze 1995 - 2003
 *
 * This file is a part of the Spectrum emulator package 'XZX-Pro', and may not be distributed.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Erik Kunze
 ********************************************************************************************************************************/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#ifdef XZX_IF1
#ifndef lint
static char rcsid[]="$Id: if1.c,v 4.48 2003/03/28 11:06:11 erik Rel $";
#endif
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>
#include "z80/z80.h"
#include "monitor/debug.h"
#include "ui/resource.h"
#include "machine/mem.h"
#include "util.h"
#include "main.h"
#include "fileio/mdr.h"
#include "ui/menu.h"
#ifdef IF1_PRINTER
#include "printer/printer.h"
#endif
#ifdef RS232
#include "rs232.h"
#endif
#include "ui/ui.h"
#include "machine/if1.h"
#define B_BUSY	0x10
#define B_DTR	0x08
#define B_GAP	0x04
#define B_SYNC	0x02
#define B_WRPROT	0x01
#define B_WAIT	0x20
#define B_CTS	0x10
#define B_ERASE	0x08
#define B_RW	0x04
#define B_CLK	0x02
#define B_DATA	0x01
#define B_TXD	0x80
#define B_NETOUT	0x01
#define B_RXD	0x01
#define B_NETIN	B_RXD
#define GAP_CLOCK_RATE	10 
typedef struct{int id;unsigned int motor:1;unsigned int pos;Cartridge*cartridge;}mdrive;
#define mdriveIsLoaded(mdp)	((mdp)->cartridge != (Cartridge *)0)
#define mdriveIsWriteProtected(mdp)	((mdp)->cartridge->readonly != 0)
#define mdriveIsReady(mdp)	((mdp)->motor != 0 && mdriveIsLoaded(mdp))
typedef struct{uns8 control;int gapClock;unsigned int gapBit:1;int preamble;unsigned int driveCount;unsigned int numDrives;mdrive*drives[8];mdrive*drive;}if1Controller;
#ifdef DEBUG
#define DEB(x)	{ if ((CFGGET(debug) & D_IF1) != 0) { x } }
#else
#define DEB(x)
#endif
static mdrive*mdriveNew(void);static int mdriveInsertCartridge(mdrive*,const char*);static void mdriveEjectCartridge(mdrive*);static void mdriveMotor(mdrive*,unsigned int);static Cartridge*cartridgeOpen(const char*);static void cartridgeClose(Cartridge*);static uns8(*if1Rs232InByte)(void);static void(*if1Rs232OutByte)(uns8);static if1Controller*if1ControllerP;void If1Init(void){if1ControllerP=Malloc(sizeof(if1Controller),"If1Init");if(LoadRomImage(CFGGET(if1Rom),IF1ROM,1)!=0){Msg(M_INFO,"Interface I emulation disabled");CFGSET(if1,-1);}else{uns16 rs232inAddr=0,rs232outAddr=0;int checksum,i;checksum=GetRomChecksum(IF1ROM,XZX_PAGESIZE);switch(checksum){case 54778:DEB(Msg(M_DEBUG,"IF1: found standard ROM v1"););rs232inAddr=0x0b9a;rs232outAddr=0x0c5a;break;case 10427:DEB(Msg(M_DEBUG,"IF1: found Ian Collier's ROM"););goto v2_rom;case 52730:DEB(Msg(M_DEBUG,"IF1: found standard ROM v2"););v2_rom:rs232inAddr=0x0b98;rs232outAddr=0x0d07;RealMemory[IF1ROM][0x16a5]=0;RealMemory[IF1ROM][0x16a6]=0;RealMemory[IF1ROM][0x16cd]=0;RealMemory[IF1ROM][0x16ce]=0;RealMemory[IF1ROM][0x16cf]=0;RealMemory[IF1ROM][0x16d0]=0;break;default:Msg(M_WARN,"unknown Interface I ROM image - checksum %u",checksum);}if(rs232inAddr!=0){RealMemory[IF1ROM][rs232inAddr]=0xed;RealMemory[IF1ROM][rs232inAddr+1]=RS232_IP_BPT;RealMemory[IF1ROM][rs232outAddr]=0xed;RealMemory[IF1ROM][rs232outAddr+1]=RS232_OP_BPT;}if1ControllerP->numDrives=NELEM(CFGGET(if1Carts));for(i=0;i<NELEM(CFGGET(if1Carts));i++){if1ControllerP->drives[i]=mdriveNew();if1ControllerP->drives[i]->id=i;if(mdriveInsertCartridge(if1ControllerP->drives[i],CFGGET(if1Carts[i]))<0){free(CFGGET(if1Carts[i]));CFGSET(if1Carts[i],(char*)0);}}if1ControllerP->drive=if1ControllerP->drives[0];}MenuFileCartridgeUpdate();If1Rs232Set(CFGGET(if1Rs232Mode));
#ifdef RS232
atexit(Rs232Close);
#endif
}int If1InsertCart(int drive){int types[2]={FILE_TYPE_MDR,FILE_TYPE_END};char*filename;int ret=-1;if((filename=UiFileSelector("Insert Cartridge",1,types))!=(char*)0){if((ret=mdriveInsertCartridge(if1ControllerP->drives[drive],filename))<0){free(filename);filename=(char*)0;}CFGSET(if1Carts[drive],filename);MenuFileCartridgeUpdate();}return ret;}void If1EjectCart(int drive){mdriveEjectCartridge(if1ControllerP->drives[drive]);CFGSET(if1Carts[drive],(char*)0);MenuFileCartridgeUpdate();}void If1ControlOutByte(uns8 val){if(val==0xe2 &&if1ControllerP->preamble==0){if1ControllerP->preamble=12;}if(val==0xee &&if1ControllerP->control==0xee){DEB(Msg(M_DEBUG,"IF1: drive count reset"););if1ControllerP->driveCount=0;val=0;}else if((val==0xec &&if1ControllerP->control==0xee)||(val==0xed &&if1ControllerP->control==0xef)){mdrive*mdp;if(if1ControllerP->driveCount>7){if1ControllerP->driveCount=0;}if((mdp=if1ControllerP->drives[8-++if1ControllerP->driveCount])!=(mdrive*)0){mdriveMotor(mdp,val==0xec ?1U:0U);}}if1ControllerP->control=val;}void If1DataOutByte(uns8 val){mdrive*mdp=if1ControllerP->drive;if(mdriveIsReady(mdp)){if(if1ControllerP->preamble==0){unsigned int pos=mdp->pos%CARTRIDGE_SECSIZE;if((pos==0&&(val&1)!=1)||(pos==15&&(val&1)!=0)){Msg(M_ERR,"microdrive write sync lost %d",mdp->pos);mdp->cartridge->readonly=1;}if(!mdriveIsWriteProtected(mdp)){mdp->cartridge->mmp[mdp->pos]=val;}if(pos==(CARTRIDGE_SECSIZE-1)||pos==14){if1ControllerP->preamble=3;}if(++(mdp->pos)==CARTRIDGE_SIZE-1){mdp->pos=0;DEB(Msg(M_DEBUG,"IF1: drive %d revolution (write)",mdp->id+1););}}else{--if1ControllerP->preamble;}}}void If1NetworkOutByte(uns8 val){USED(val);}uns8 If1ControlInByte(void){uns8 res=0xe0 |B_GAP|B_SYNC|B_WRPROT;mdrive*mdp=if1ControllerP->drive;if(mdriveIsReady(mdp)){unsigned int pos=mdp->pos%CARTRIDGE_SECSIZE;if(pos==0||pos==15){res&=~B_SYNC;}else{if(++(mdp->pos)==CARTRIDGE_SIZE-1){mdp->pos=0;DEB(Msg(M_DEBUG,"IF1: drive %d revolution (sync)",mdp->id+1););}}if(mdriveIsWriteProtected(mdp)){res&=~B_WRPROT;}if(++if1ControllerP->gapClock%GAP_CLOCK_RATE==0){if1ControllerP->gapBit^=1;}if(if1ControllerP->gapBit==0){res&=~B_GAP;}}return res;}uns8 If1DataInByte(void){uns8 res=0xff;mdrive*mdp=if1ControllerP->drive;if(mdriveIsReady(mdp)){res=mdp->cartridge->mmp[mdp->pos];if(++(mdp->pos)==CARTRIDGE_SIZE-1){mdp->pos=0;DEB(Msg(M_DEBUG,"IF1: drive %d revolution (read)",mdp->id+1););}}return res;}uns8 If1NetworkInByte(void){return 0xff;}void If1Rs232Set(int mode){if1Rs232InByte=IoDummyInByte;if1Rs232OutByte=IoDummyOutByte;
#ifdef RS232
if(mode==1){if(Rs232Open()==0){if1Rs232InByte=Rs232InByte;if1Rs232OutByte=Rs232OutByte;}else{mode=0;}}else{Rs232Close();}
#endif
#ifdef IF1_PRINTER
if(mode==2){if1Rs232OutByte=If1SerialOutByte;}
#endif
CFGSET(if1Rs232Mode,mode);}void If1Rs232InByte(void){uns8 res=if1Rs232InByte();DEB(Msg(M_DEBUG,"IF1: RS232 read <%02x>",(unsigned int)res););A=res;RET();}void If1Rs232OutByte(void){uns8 val=A;DEB(Msg(M_DEBUG,"IF1: RS232 write <%02x>",(unsigned int)val););if1Rs232OutByte(val);RET();}static mdrive*mdriveNew(void){return(Malloc(sizeof(mdrive),"mdriveNew"));}static int mdriveInsertCartridge(mdrive*mdp,const char*filename){if(mdp->cartridge!=(Cartridge*)0){mdriveEjectCartridge(mdp);}if((mdp->cartridge=cartridgeOpen(filename))==(Cartridge*)0){return-1;}mdp->pos=0;return 0;}static void mdriveEjectCartridge(mdrive*mdp){cartridgeClose(mdp->cartridge);mdp->cartridge=(Cartridge*)0;}static void mdriveMotor(mdrive*mdp,unsigned int state){DEB(Msg(M_DEBUG,"IF1: drive %d motor %s",mdp->id+1,state==0?"off":"on"););if1ControllerP->drive=mdp;if(state!=0){if1ControllerP->gapClock=0;if1ControllerP->gapBit=0;if1ControllerP->preamble=0;if(mdriveIsLoaded(mdp)||(CFGGET(smartAttach)!=0&&If1InsertCart(mdp->id)==0)){mdp->pos/=CARTRIDGE_SECSIZE;mdp->pos*=CARTRIDGE_SECSIZE;}}mdp->motor=state;}static Cartridge*cartridgeOpen(const char*filename){assert(GetFileType(filename)==FILE_TYPE_MDR);return(MdrOpen(filename));}static void cartridgeClose(Cartridge*cp){assert(cp!=(Cartridge*)0);Munmap(cp->mmp,cp->mmsize);free((void*)cp->filename);free(cp);}
#endif 

