/*********************************************************************************************************************************
 * Copyright (c) Erik Kunze 1995 - 2003
 *
 * This file is a part of the Spectrum emulator package 'XZX-Pro', and may not be distributed.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Erik Kunze
 ********************************************************************************************************************************/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#ifndef lint
static char rcsid[]="$Id: sound_alsa.c,v 4.9 2003/04/25 22:51:57 erik Rel $";
#endif
#include <sys/types.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>
#define ALSA_PCM_NEW_HW_PARAMS_API
#define ALSA_PCM_NEW_SW_PARAMS_API
#include ASOUNDLIB_H_PATH
#include "monitor/debug.h"
#include "ui/resource.h"
#include "util.h"
#include "sound/sound.h"
#ifdef SND_LIB_VERSION
#if SND_LIB_VERSION >= (5 << 8)
#if SND_LIB_VERSION >= (9 << 8)
#define NEW_INTERFACE
#endif
#else
#error "versions prior 0.5 are not supported anymore"
#endif
#endif
#ifdef DEBUG
#define DEB(x)	{ if ((CFGGET(debug) & D_SOUND) != 0) { x } }
#else
#define DEB(x)
#endif
static snd_pcm_t*pcmHandle=(snd_pcm_t*)0;
#ifdef DEBUG
static unsigned int numberOfWrites,numberOfUnderruns;
#endif
#ifdef NEW_INTERFACE
int OsdOpenAudioDevice(const char*device,int*sample,int*stereo){int card,dev;char name[32];snd_ctl_t*ctlHandle;snd_ctl_card_info_t*info;snd_pcm_info_t*pcminfo;snd_pcm_hw_params_t*hwParams;snd_pcm_sw_params_t*swParams;unsigned int sampleMin,sampleMax,sampleCurrent,channelsMax,bufferTime;snd_pcm_uframes_t bufferSize,frames;int err,i;USED(device);card=-1;if((err=snd_card_next(&card))<0||card<0){Msg(M_WARN,"no sound cards detected");return-2;}(void)sprintf(name,"hw:%d",card);if((err=snd_ctl_open(&ctlHandle,name,0))<0){Msg(M_ERR,"couldn't open control device\n  -> reason: %s",snd_strerror(err));goto error;}snd_ctl_card_info_alloca(&info);if((err=snd_ctl_card_info(ctlHandle,info))<0){Msg(M_ERR,"couldn't read card info\n  -> reason: %s",snd_strerror(err));goto error;}snd_pcm_info_alloca(&pcminfo);dev=-1;if(snd_ctl_pcm_next_device(ctlHandle,&dev)<0||dev<0){Msg(M_WARN,"no sound cards detected");snd_ctl_close(ctlHandle);return-2;}snd_pcm_info_set_device(pcminfo,(unsigned int)dev);snd_pcm_info_set_subdevice(pcminfo,0);snd_pcm_info_set_stream(pcminfo,SND_PCM_STREAM_PLAYBACK);if((err=snd_ctl_pcm_info(ctlHandle,pcminfo))<0){Msg(M_ERR,"unable to obtain name\n  -> reason: %s",snd_strerror(err));goto error;}DEB(Msg(M_DEBUG,"SND: soundcard        %s",snd_ctl_card_info_get_name(info)););snd_ctl_close(ctlHandle);if((err=snd_pcm_open(&pcmHandle,name,SND_PCM_STREAM_PLAYBACK,SND_PCM_NONBLOCK))<0){Msg(M_ERR,"couldn't open device %d, card %d for writing\n  -> reason: %s",dev,card,snd_strerror(err));return(err==-EBUSY||err==-EAGAIN?-1:-2);}
#ifdef SOUND_NONBLOCKING
i=1;
#else
i=0;
#endif
if((err=snd_pcm_nonblock(pcmHandle,i))<0){Msg(M_ERR,"couldn't set %sblocking mode\n  -> reason: %s",i==0?"non":"",snd_strerror(err));goto error;}snd_pcm_hw_params_alloca(&hwParams);if((err=snd_pcm_hw_params_any(pcmHandle,hwParams))<0){Msg(M_ERR,"broken configuration for playback\n  -> reason: %s",snd_strerror(err));goto error;}(void)snd_pcm_hw_params_get_rate_min(hwParams,&sampleMin,0);(void)snd_pcm_hw_params_get_rate_max(hwParams,&sampleMax,0);(void)snd_pcm_hw_params_get_channels_max(hwParams,&channelsMax);DEB(Msg(M_DEBUG,"SND: min sample rate  %u",sampleMin);Msg(M_DEBUG,"SND: max sample rate  %u",sampleMax);Msg(M_DEBUG,"SND: channels         %u",channelsMax););if((err=snd_pcm_hw_params_set_access(pcmHandle,hwParams,SND_PCM_ACCESS_RW_INTERLEAVED))<0){Msg(M_ERR,"interleaved access not available\n  -> reason: %s",snd_strerror(err));goto error;}if((err=snd_pcm_hw_params_set_format(pcmHandle,hwParams,SND_PCM_FORMAT_U8))<0){Msg(M_ERR,"couldn't set sample resolution to 8 bit\n  -> reason: %s",snd_strerror(err));goto error;}*stereo=channelsMax>1;if((err=snd_pcm_hw_params_set_channels(pcmHandle,hwParams,*stereo<<1))<0){Msg(M_ERR,"couldn't set %s output",stereo?"stereo":"mono");goto error;}if(*sample>(int)sampleMax){*sample=(int)sampleMax;}else if(*sample<(int)sampleMin){*sample=(int)sampleMin;}sampleCurrent=(unsigned int)*sample;if((err=snd_pcm_hw_params_set_rate_near(pcmHandle,hwParams,&sampleCurrent,0))<0){Msg(M_PERR,"couldn't set sample rate to %d Hz",*sample);goto error;}if(sampleCurrent!=(unsigned int)*sample){Msg(M_WARN,"using different sample rate (requested %uHz, got %dHz)",sampleCurrent,*sample);*sample=(int)sampleCurrent;}bufferTime=100000;if((err=snd_pcm_hw_params_set_buffer_time_near(pcmHandle,hwParams,&bufferTime,0))<0){Msg(M_ERR,"couldn't set buffer time %u for playback\n  -> reason: %s",bufferTime,snd_strerror(err));goto error;}frames=1;if((err=snd_pcm_hw_params_set_period_size_near(pcmHandle,hwParams,&frames,0))<0){Msg(M_ERR,"couldn't set period size %lu for playback\n  -> reason: %s",frames,snd_strerror(err));goto error;}if((err=snd_pcm_hw_params(pcmHandle,hwParams))<0){Msg(M_ERR,"unable to set hardware paramaters\n  -> reason: %s",snd_strerror(err));goto error;}(void)snd_pcm_hw_params_get_buffer_size(hwParams,&bufferSize);snd_pcm_sw_params_alloca(&swParams);if((err=snd_pcm_sw_params_current(pcmHandle,swParams))<0){Msg(M_ERR,"unable to determine current settings for playback\n  -> reason: %s",snd_strerror(err));goto error;}if((err=snd_pcm_sw_params_set_sleep_min(pcmHandle,swParams,0))<0||(err=snd_pcm_sw_params_set_xfer_align(pcmHandle,swParams,(snd_pcm_uframes_t)1))<0||(err=snd_pcm_sw_params_set_avail_min(pcmHandle,swParams,1))<0||(err=snd_pcm_sw_params_set_start_threshold(pcmHandle,swParams,1))<0||(err=snd_pcm_sw_params_set_stop_threshold(pcmHandle,swParams,bufferSize))<0||(err=snd_pcm_sw_params(pcmHandle,swParams))<0){Msg(M_ERR,"unable to set software paramaters\n  -> reason: %s",snd_strerror(err));goto error;}if((err=snd_pcm_prepare(pcmHandle))<0){Msg(M_ERR,"unable to prepare channel\n  -> reason: %s",snd_strerror(err));goto error;}return 0;error:if(ctlHandle!=(snd_ctl_t*)0){(void)snd_ctl_close(ctlHandle);}if(pcmHandle!=(snd_pcm_t*)0){(void)snd_pcm_close(pcmHandle);}pcmHandle=(snd_pcm_t*)0;return-1;}
#else
int OsdOpenAudioDevice(const char*device,int*sample,int*stereo){snd_pcm_channel_info_t info;snd_pcm_channel_params_t params;snd_pcm_channel_setup_t setup;const char*cardname;int cards,card=0,dev=0,err;int i,j;USED(device);if((cards=snd_cards())<=0){Msg(M_WARN,"no sound cards detected");return-2;}DEB(Msg(M_DEBUG,"SND: soundcards       %d",cards););if((err=snd_card_get_name(0,(char**)&cardname))<0){Msg(M_ERR,"unable to obtain name\n  -> reason: %s",snd_strerror(err));return-2;}DEB(Msg(M_DEBUG,"SND: soundcard        %s",cardname););free((void*)cardname);if((err=snd_pcm_open(&pcmHandle,card,dev,SND_PCM_OPEN_PLAYBACK|SND_PCM_OPEN_NONBLOCK))<0){Msg(M_ERR,"couldn't open device %d, card %d for writing\n  -> reason: %s",dev,card,snd_strerror(err));return(err==-EBUSY||err==-EAGAIN?-1:-2);}
#ifdef SOUND_NONBLOCKING
i=0;
#else
i=1;
#endif
(void)snd_pcm_nonblock_mode(pcmHandle,i);memset(&info,0,sizeof(info));info.channel=SND_PCM_CHANNEL_PLAYBACK;if((err=snd_pcm_channel_info(pcmHandle,&info))<0){Msg(M_ERR,"couldn't get audio settings\n  -> reason: %s",snd_strerror(err));goto error;}DEB(Msg(M_DEBUG,"SND: min sample rate  %u",info.min_rate);Msg(M_DEBUG,"SND: max sample rate  %u",info.max_rate);Msg(M_DEBUG,"SND: channels         %d",info.max_voices););if(*sample>info.max_rate){*sample=info.max_rate;}else if(*sample<info.min_rate){*sample=info.min_rate;}memset(&params,0,sizeof(params));params.channel=SND_PCM_CHANNEL_PLAYBACK;params.mode=SND_PCM_MODE_STREAM;params.start_mode=SND_PCM_START_DATA;params.stop_mode=SND_PCM_STOP_STOP;params.format.format=SND_PCM_SFMT_U8;params.format.rate=*sample;params.format.voices=info.max_voices>1?2:1;params.format.interleave=1;i=params.format.rate*params.format.voices;i>>=2;if(i<16){i=16;}for(j=1;j<i;j<<=1);params.buf.stream.queue_size=j>>1;params.buf.stream.fill=SND_PCM_FILL_NONE;DEB(Msg(M_DEBUG,"SND: queue size       %u",params.buf.stream.queue_size););if((err=snd_pcm_channel_params(pcmHandle,&params))<0){Msg(M_ERR,"couldn't set 8 bit, %d Hz, %s\n  -> reason: %s",params.format.rate,params.format.voices==2?"stereo":"mono",snd_strerror(err));goto error;}*stereo=params.format.voices==2;if((err=snd_pcm_channel_prepare(pcmHandle,SND_PCM_CHANNEL_PLAYBACK))<0){Msg(M_ERR,"unable to prepare channel\n  -> reason: %s",snd_strerror(err));goto error;}memset(&setup,0,sizeof(setup));setup.mode=SND_PCM_MODE_STREAM;setup.channel=SND_PCM_CHANNEL_PLAYBACK;if((err=snd_pcm_channel_setup(pcmHandle,&setup))<0){Msg(M_ERR,"unable to obtain setup\n  -> reason: %s",snd_strerror(err));goto error;}
#ifdef DEBUG
numberOfWrites=0;numberOfUnderruns=0;
#endif
return 0;error:(void)snd_pcm_close(pcmHandle);pcmHandle=(snd_pcm_t*)0;return-1;}
#endif
#ifdef NEW_INTERFACE
void OsdCloseAudioDevice(void){assert(pcmHandle!=(snd_pcm_t*)0);(void)snd_pcm_drain(pcmHandle);(void)snd_pcm_close(pcmHandle);pcmHandle=(snd_pcm_t*)0;DEB(Msg(M_DEBUG,"SND: writes           %u",numberOfWrites);Msg(M_DEBUG,"SND: underruns        %u",numberOfUnderruns););}
#else
void OsdCloseAudioDevice(void){assert(pcmHandle!=(snd_pcm_t*)0);(void)snd_pcm_playback_flush(pcmHandle);(void)snd_pcm_close(pcmHandle);pcmHandle=(snd_pcm_t*)0;DEB(Msg(M_DEBUG,"SND: writes           %u",numberOfWrites);Msg(M_DEBUG,"SND: underruns        %u",numberOfUnderruns););}
#endif
#ifdef NEW_INTERFACE
int OsdDrainAudioDevice(void){return(snd_pcm_drain(pcmHandle));}
#else
int OsdDrainAudioDevice(void){return(snd_pcm_playback_drain(pcmHandle));}
#endif
#ifdef SOUND_NONBLOCKING
#ifdef NEW_INTERFACE
int OsdGetAudioFreespace(void){snd_pcm_status_t*status;snd_pcm_status_alloca(&status);return(snd_pcm_status(pcmHandle,status)<0?-1:(int)snd_pcm_status_get_avail(status));}
#else
int OsdGetAudioFreespace(void){snd_pcm_channel_status_t status;memset(&status,0,sizeof(status));status.channel=SND_PCM_CHANNEL_PLAYBACK;return(snd_pcm_channel_status(pcmHandle,&status)<0?-1:status.free);}
#endif
#endif
#ifdef NEW_INTERFACE
int OsdPlayAudioBuffer(unsigned char*buf,int len){int written,remain;
#ifdef DEBUG
numberOfWrites++;
#endif
for(written=0,remain=len;remain>0;buf+=written,remain-=written){do{written=(int)snd_pcm_writei(pcmHandle,buf,(size_t)remain);}while(written==-EINTR);if(written<0){break;}}if(written<0&&written!=-EAGAIN){if(written==-EPIPE){
#ifdef DEBUG
numberOfUnderruns++;
#endif
if((written=snd_pcm_prepare(pcmHandle))<0){Msg(M_ERR,"couldn't recover from underrun\n  -> reason: %s",snd_strerror(written));}}else if(written==-ESTRPIPE){snd_pcm_status_t*status;snd_pcm_status_alloca(&status);if(snd_pcm_status(pcmHandle,status)==0&&snd_pcm_status_get_state(status)==SND_PCM_STATE_XRUN){written=snd_pcm_prepare(pcmHandle);}}else{Msg(M_ERR,"couldn't write to audio device\n  -> reason: %s",snd_strerror(written));}return written;}return len-remain;}
#else
int OsdPlayAudioBuffer(unsigned char*buf,int len){int written,remain;
#ifdef DEBUG
numberOfWrites++;
#endif
for(written=0,remain=len;remain>0;buf+=written,remain-=written){do{written=(int)snd_pcm_write(pcmHandle,buf,(size_t)remain);}while(written==-EINTR);if(written<0){break;}}if(written<0&&written!=-EAGAIN){if(written==-EPIPE||written==-EBADFD){snd_pcm_channel_status_t status;memset(&status,0,sizeof(status));status.channel=SND_PCM_CHANNEL_PLAYBACK;if(snd_pcm_channel_status(pcmHandle,&status)==0&&status.status==SND_PCM_STATUS_UNDERRUN){
#ifdef DEBUG
numberOfUnderruns++;
#endif
written=snd_pcm_channel_prepare(pcmHandle,SND_PCM_CHANNEL_PLAYBACK);}}else{Msg(M_ERR,"couldn't write to audio device\n  -> reason: %s",snd_strerror(written));}return written;}return len-remain;}
#endif

